#ifndef DIAGONALS_HPP
#define DIAGONALS_HPP

#include <algorithm>
#include <array>
#include <iostream>
#include <vector>

#include "global.hpp"
#include "integer.hpp"
#include "square.hpp"
#include "fstack.hpp"

// Diag positions are 0 indexed
struct BitwiseDiag {
  __uint128_t low;
  std::array<unsigned int, N> vals;

  BitwiseDiag();

  unsigned int get(unsigned int index) const;
  void set(unsigned int i, unsigned int val);
  bool conflict(const BitwiseDiag& d) const;
  bool conflict(const BitwiseDiag* d) const;
  bool conflict(const __uint128_t& cell) const;
  bool intersect(const BitwiseDiag& d) const;
  bool intersect(const BitwiseDiag* d) const;
  void show();
  void print(std::ostream& stream) const;
};

struct SimpleDiag {
  std::array<unsigned int, N> diagonal;
  bool active;

  SimpleDiag();

  unsigned int get(unsigned int index) const;
  void set(unsigned int i, unsigned int val);
  bool conflict(const SimpleDiag& d);
  bool intersect(const SimpleDiag& d);
  void show();
  void print(std::ostream& stream) const;
};

// typedef SimpleDiag Diag;
typedef BitwiseDiag Diag;

std::ostream& operator << (std::ostream& stream, const Diag& diag);

std::ostream& operator << (std::ostream& stream,
  const std::vector<Diag>& diags);

std::ostream& operator << (std::ostream& stream,
  const std::vector<Diag*>& diags);

std::ostream& operator << (std::ostream& stream,
  const FStack<Diag*>& diags);

class DiagGenerator {
private:
  // Const referencial data
  const std::vector<Square>& m_squares;


  // Keeps the number of necessary frequencyCounts
  std::vector<unsigned int> m_freqCounts;
  std::array<unsigned int, 2> m_oCounts;

  // Update variables
  std::vector<unsigned int> m_freqHits;
  unsigned int m_oHits;

  // Diagonal for storage
  Diag m_diag;
  std::array<std::vector<Diag>, 2> m_diagonals;

  // Helper methods for the generation.
  bool hasFrequencySpace(uint_mod<2> square, unsigned int row,
    unsigned int col);
  void updateFrequencies(uint_mod<2> square, unsigned int row,
    unsigned int col);
  void restoreFrequencies(uint_mod<2> square, unsigned int row,
    unsigned int col);
  bool symbolMatch(unsigned int square, unsigned int row, unsigned int col);
  void generate(const uint_mod<2> square,
    const std::vector<unsigned int>& available);
public:
  DiagGenerator(const std::vector<Square>& squares);
  void generate();

  std::vector<Diag>& getSDiagonals();
  std::vector<Diag>& getTDiagonals();

  void print(std::ostream& stream) const;
};

std::ostream& operator << (std::ostream& stream, const DiagGenerator& gen);
/**
 * \brief Generates pairs of diagonalsthat constitute possible solutions to the MOLS satisfying the frequency squares.
 *
 * @param mols A set of partially completed mols to be completed
 * @param fSquares A set of frequency squares
 * @param diagonalPairs Contains the restults of attemting to complete the diagonals.
 *
 */
// void generateSquareDiagonals(const uint_mod<2> type, const MOLSPair& mols, const FrequencySquares& fSquares, std::vector<Diag>& diagonals);

#endif // DIAGONALS_HPP
