// AUTHOR: Michael Gill
// School or Mathematics
// Monash University
// Clayton, Victoria, AU 3800
// email:  Michael.Gill@Monash.edu
//
// A set of program wide definitions and typename declarations. All header files
// should include this file to be ensure project consistency.
// Copyright (C) 2019  Michael Gill
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
#ifndef LATIN10_IO_FILE_HPP_
#define LATIN10_IO_FILE_HPP_

#include <vector>
#include <assert.h>

#include "definitions.hpp"
#include "io/decoder.hpp"
#include "io/encoder.hpp"
#include "util/list.hpp"
#include "template/template.hpp"

enum encoding {
  NUMERIC,
  ALPHA,
  BITPACKED,
  FACTORIAL
};

struct TemplateOptions {
  size_t squares = 0;
  std::vector<size_t> gSizes;
  std::array<size_t, 6> bSizes;
  size_t rIdx = 0;
  size_t cIdx = 1;
  bool valid;
};

struct FileOptions {
  size_t squares = 2;
  encoding enc = FACTORIAL;
  std::vector<size_t> gSizes;
  std::array<size_t, 6> bSizes;
  size_t rIdx;
  size_t cIdx;
  bool valid;
  size_t skip = 0;
  size_t offset = 0;
  size_t ms1 = 0;
  size_t ms2 = 1;
};

class FileReader {
private:
  Decoder* d;
  std::istream& stream;
  FileOptions& options;
  std::string sqID;
public:
  FileReader(std::istream& stream, FileOptions& opt);
  ~FileReader();
  std::string nextID() const;
  bool hasNext() const;
  void read(std::vector<Square>& squares);
  std::string readAsString();
  void trackNext();
  void track(size_t t);
};

class FileWriter {
private:
  Encoder* e;
  std::ostream& stream;
  FileOptions& options;
  std::string idPrefix;
public:
  FileWriter(std::ostream& stream, FileOptions& opt);
  FileWriter(std::ostream& stream, FileOptions& opt, std::string idPrefix);
  ~FileWriter();
  void write(const std::vector<Square>& squares, const std::string& id);
  void writeAsString(const std::string& squares, const std::string& id);
};

class TemplateReader {
private:
  std::istream& stream;
  TemplateOptions& options;
  std::string template_id;
public:
  TemplateReader(std::istream& input, TemplateOptions& opts);
  ~TemplateReader();
  template <bool HotStart = false>
  void read(Template&);
  bool hasNext() const;
  std::string nextID() const ;
  void skip(size_t i);
};

class TemplateWriter {
private:
  std::ostream& stream;
  TemplateOptions& options;
  std::string idPrefix;
public:
  TemplateWriter(std::ostream& stream, TemplateOptions& opt);
  TemplateWriter(std::ostream& stream, TemplateOptions& opt, std::string idPrefix);
  ~TemplateWriter();
  void write(const std::vector<Square>& squares, const std::string& id);
};

template <bool HotStart>
void TemplateReader::read(Template& temp) {
  for (size_t row = 0; row < N; row++) {
    for (size_t col = 0; col < N; col++) {
      for (size_t sym = 0; sym < temp.getWeight(); sym++) {
        temp.clear(row, col, sym);
      }
    }
  }

  std::string line;
  for (size_t row = 0; row < N; row++) {
    std::getline(stream, line);
    for (size_t col = 0; col < N; col++) {
      size_t val = hexctoi(line[col]);
      if (val > 0) {
        for (size_t sym = 0; sym < MAX_SYMB; sym++) {
          if (val & (1 << sym)) {
            temp.set(row, col, sym);
          }
        }
      }
    }
  }


  if constexpr (HotStart) {

    size_t orbits = 0;
    stream >> orbits;

    temp.hs_reserve(orbits);

    for (size_t i = 0; i < orbits; i++) {

      for (size_t j = 0; j < 2; j++) {
        size_t count = 0;
        stream >> count;

        size_t square = 0;
        stream >> square;

        if (count > 0) {
          temp.hs_orbit_reserve(count, square, i);
        }
        for (size_t k = 0; k < count; k++) {
          stream >> line;
          temp.add_hs_diag(Diag(line), square, i);
        }
      }
    }
  }

  std::getline(stream, template_id);
}

#endif // LATIN10_IO_FILE_HPP_
