// AUTHOR: Michael Gill
// School or Mathematics
// Monash University
// Clayton, Victoria, AU 3800
// email:  Michael.Gill@Monash.edu
//
// Integer manipulations required for encoding and decoding LS.
// Copyright (C) 2019  Michael Gill
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
#ifndef LATIN10_NUMBER_INTEGER_HPP_
#define LATIN10_NUMBER_INTEGER_HPP_

#include <vector>
#include <string>
#include <numeric>
#include <sstream>
#include <list>

static_assert(sizeof(size_t) == 8, "Integer library requires integer width of \
  at least 36-bits");

#include "definitions.hpp"

#define B_MASK 63
#define BLOCK_S 6
#define B_DELIM 126
#define MID 3
#define HIGH 6

// Factorial of numbers up to N are required. This can easily be stored for
// quick lookup.
const std::array<size_t, N+1> factorial = {
  1,
  1,
  2,
  6,
  24,
  120,
  720,
  5040,
  40320,
  362880,
  3628800
};

// Let r be a string encoding consistent with R of k bytes.
//
// Returns the value x encoded by the string r. Each byte of r represents,
// 6-bits of x, bigendian aligned. Each 6-bit blocks is obtained from x by
// subtracting 63 from the ordinal value of the char.
//
// Example:
//   r = "]Gw" = 01011101 01000111 01110111
//                  93        71     119
//
//   x = 011110 000100 111000 = 123448
//         30      8     56
size_t R_inv(const std::string& r, size_t blocks);

// Let x be an integer consiting of k blocks of 6-bits.
//
// The R(x, k) returns a string containing the integer x encoded into k bytes.
// Let x[i] be the block i. The the byte at i in the return string is x[i]+63
// Each byte takes a value between 63 and 127. This places each byte of the
// string in the human readable ASCII values.
//
// Example:
//   x = 215940 : 110100 101110 000100
//                  52     46      4
//   R(x, 3) =    [115,   109,    67]
std::string R(size_t x, size_t blocks);

// Let x be an integer consiting of 1 or 3n blocks of 6-bits.
//
// Returns a string representing x encoded using R(x,k) with the a prefix
// consisting of repetitions of the character '~' in the following way.
//
// 1) If no prefix is present, then pack(x) consists of a single byte.
//
// 2) If teh prefix has length l, then pack(x) consists of 4l bytes with the
//    suffix of length 3l an R(x, 3l) encoded integer.
//
// Example:
//   x = 215940 : 110100 101110 000100
//   pack(x) = "~smC"
std::string pack(size_t x);

// Let r be a string encoding an integer consistent with pack(x)
//
// Returns the value x encoded by the string
//
// Example:
//   r = "~]Gw"
//   unpack(r) = 011110 001000 111000
//                 30      8     56
//   unpack(r) = 123448
size_t unpack(const std::string& r);

// Calculate the permutation of symbols from 1..N at postion i in a
// lexicographic ordering of all permutation of order N.
//
// Example:
//   Let N = 3, then there are 6 permutations. given in lexicographic order
//   they are: 123, 132, 213, 231, 312, 321
//
//   The permutation at postition 3 is 231.
const Permutation permFromPosition(size_t position);

// A derangement of a permutation, pi, is a reordering of the symbols in pi,
// kappa, such that for each position i, pi[i] != kappa[i].
//
// Returns a derangement, d, of the identity permutation of order N such that
// the lexicographic position of d in the set of permutations of order N is a
// minimum.
//
// Constexpr is fiddly. There is almost certainly a better way of doing the
// initialiser list.
//
// TODO(Michael): Improve initialiser list syntax.
constexpr Permutation getMinimalDerangement() {
  return {2,1,4,3,6,5,8,7,10,9};
}

// Templated functions

// Determine the lexicographic position of the permutation of order N starting
// at the forward iterator 'start' under some mapping theta.
template <typename ForwardIterator, typename List>
size_t lexPermutationPosition(ForwardIterator start, List map) {
  size_t position = 0;

  // Seen is 0 indexed. Symbols start at 1.
  bool seen[N];
  for (size_t i = 0; i < N; i++) {
    seen[i] = false;
  }

  // The number of branches skipped due to choosing a symbol x at position i of
  // your permutation is the number of unseen values smaller that x that occure
  // before position i, count, plus the number of symbols that are remaining.
  // This gives count * (n - (i+1)).
  ForwardIterator end = start;
  for (size_t k = 0; k < N-1; k++) end++;
  for (ForwardIterator it = start; it != end; it++) {
    // Calculate the number of branches we skipped by this choice.
    size_t count = 0;
    for (unsigned int sym = 1; sym < map[*it-1]; sym++) { if (!seen[sym-1]) count++; }

    position += count*factorial[(N - (std::distance(start, it) + 1))];

    seen[map[*it-1]-1] = true;
  }

  return position;
}

// Determine the lexicographic position of the permutation of order N starting
// at the forward iterator 'start', using the trivial mapping.
template <typename ForwardIterator>
size_t lexPermutationPosition(ForwardIterator start) {
  std::vector<size_t> map(N);
  std::iota(map.begin(), map.end(), 1);

  return lexPermutationPosition(start, map);
}

// Inherited from gddSolver
static_assert(sizeof(uint_fast64_t) >= 8, "uint_fast64_t is of an inconsistent size.");

template <unsigned int T>
struct uint_mod {
  unsigned int val;

  uint_mod(unsigned int input) {
    val = input % T;
  }

  void operator=(unsigned int i) {
    val = i % T;
  }

  friend uint_mod<T> operator+(uint_mod<T> lhs, const unsigned int& rhs) {
    lhs = lhs.val + rhs;
    return lhs;
  }

  explicit operator unsigned int() const {
    return val;
  }
};

unsigned int popCount(uint_fast64_t a);

size_t hexctoi(const char c);
char itohexc(const size_t i);

#endif // LATIN10_NUMBER_INTEGER_HPP_
