// AUTHOR: Michael Gill
// School or Mathematics
// Monash University
// Clayton, Victoria, AU 3800
// email:  Michael.Gill@Monash.edu
//
// Functions for dealing with answers relating to transversals.
// Example: [Find the number of common transversals,
//           Find the largest set of common disjoint transversals]
// Copyright (C) 2019  Michael Gill
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "io/cloptions.hpp"

CLOption::CLOption() {}

CLOption::CLOption(const std::string &key, const std::string &description,
                   TYPE t_def)
    : key(key), description(description), t_def(t_def) {}

TYPE CLOption::getType() { return t_def; }

CLResults::CLResults(CLOptions &options) : options(options) {}

std::optional<std::any> CLResults::operator[](const char *value) {
  if (result_map.count(options[value])) {
    return result_map[options[value]];
  }

  return std::nullopt;
}

bool CLResults::is_valid() { return valid; }

CLResults::operator bool() { return is_valid(); }

std::string CLOption::getDesc() { return description; }

CLOptions::CLOptions() { addOption("help,h", "Show this help dialog"); }

CLOptions::CLOptions(const std::string &name) : name(name) {
  addOption("help,h", "Show this help dialog");
}

bool CLOptions::addOption(const std::string &key, const std::string &desc) {
  return addOption(key, desc, NONE);
}

bool CLOptions::addOption(const std::string &key, const std::string &desc,
                          TYPE t_def) {
  std::regex re("^(\\S+),(\\S)$");
  std::smatch matches;
  if (std::regex_match(key, matches, re)) {
    for (size_t i = 1; i < matches.size(); i++) {
      key_map.insert({matches[i], matches[0]});
    }
    option_map.insert({key, CLOption(key, desc, t_def)});
    return true;
  }

  return false;
}

CLResults CLOptions::parse(const int argc, const char **argv) {
  bool opt_mode = true;
  CLResults results(*this);

  int pos = 1;
  while (pos < argc) {
    std::string value = argv[pos++];
    // If is command
    if (opt_mode && value[0] == '-') {
      size_t offset = (value[1] == '-') ? 2 : 1;
      value = value.substr(offset, value.size());
      value = (*this)[value];
      if (!getOption(value)) {
        results.valid = false;
        return results;
      }

      if (getOption(value)->getType() == NONE) {
        results.result_map.insert({value, true});
      } else {
        if (pos == argc) {
          results.valid = false;
          return results;
        }

        if (getOption(value)->getType() == INTEGER) {
          try {
            results.result_map.insert({value, std::stoi(argv[pos++])});
          } catch (std::exception &e) {
            results.valid = false;
            return results;
          }
        } else if (getOption(value)->getType() == DOUBLE) {
          try {
            results.result_map.insert({value, std::stod(argv[pos++])});
          } catch (std::exception &e) {
            results.valid = false;
            return results;
          }
        } else if (getOption(value)->getType() == STRING) {
          results.result_map.insert({value, std::string(argv[pos++])});
        }
      }
    } else {
      if (opt_mode) {
        opt_mode = false;
        results.positional.reserve(argc - pos);
      }
      results.positional.push_back(argv[pos++]);
    }
  }

  results.valid = true;

  return results;
}

const std::string &CLOptions::operator[](const std::string &value) {
  return key_map[value];
}

std::optional<CLOption> CLOptions::getOption(const std::string &key) {
  if (option_map.count(key)) {
    return option_map[key];
  }

  return {};
}

std::string CLOptions::usage() {
  std::stringstream ss;
  ss << name << "\n" << USAGE_HEADER << "\n";
  for (auto option : option_map) {
    ss << "\t" << option.first << "\t" << option.second.getDesc() << "\n\n";
  }

  return ss.str();
}
