// AUTHOR: Michael Gill
// School or Mathematics
// Monash University
// Clayton, Victoria, AU 3800
// email:  Michael.Gill@Monash.edu
//
// Implementation of the decoders. For encoding specifications see the manual
// io section
// Copyright (C) 2019  Michael Gill
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "io/decoder.hpp"

#include "assert.h"

Decoder::~Decoder() {}

void NumericDecoder::decode(std::istream &stream, Square &square) {
  unsigned short val = 0;
  for (auto& s_val : square) {
    stream >> val;
    s_val = val;
  }
}

void AlphaDecoder::decode(std::istream &stream, Square &square) {
  for (size_t row = 0; row < N; row++) {
    std::string line;
    std::getline(stream, line);

    for (unsigned int col = 0; col < N; col++) {
      square[row*N + col] = line[col] - ('a' - 1);
    }
  }
}

void BitDecoder::decode(std::istream &stream, Square &square) {
  for (size_t i = 0; i < N; i++) {
    square[i] = i+1;
  }

  std::string str;
  getline(stream, str);

  size_t begin = 0;
  for (size_t i = 1; i < N-1; i++) {
    // Extract the row
    size_t count = 0;
    while (str[begin+count] == B_DELIM) {
      count++;
    }
    if (count > 0) {
      count = 4*count;
    } else {
      count++;
    }
    std::string part = str.substr(begin, count);

    // Insert the row
    size_t remaining = factorial[N];
    size_t packed_int = unpack(part);
    for (size_t j = 0; j < N-1; j++) {
      size_t hb = (packed_int >> HALF_BYTE_S*(N-2-j)) & HALF_BYTE_M;
      square[i*N + j] = hb;
      remaining /= hb;
    }
    square[i*N + N-1] = remaining;

    begin += count;
  }

  for (size_t col = 0; col < N; col++) {
    size_t remaining = factorial[N];
    for (size_t row = 0; row < N-1; row++) {
      remaining /= square[row*N + col];
    }
    square[(N-1)*N + col] = remaining;
  }
}

// Factorial decoder pulls directly from istream using the overloaded "<<"
// operator. This is opposed to the getline command. This choice is because
// the "<<" operator will work if the istream is lead with a "\n"... Possible
// if you use the "<<" operator in a driver function.
void FactorialDecoder::decode(std::istream &stream, Square &square) {
  if (minPos == 0) {
    minPos = lexPermutationPosition(getMinimalDerangement().begin());
  }

  for (size_t i = 0; i < N; i++) {
    square[i] = i+1;
  }

  std::string str;
  std::getline(stream, str);

  size_t begin = 0;
  for (size_t i = 1; i < N-1; i++) {
    // Extract the row
    size_t count = 0;
    while (str[begin+count] == B_DELIM) {
      count++;
    }

    if (count > 0) {
      count = 4*count;
    } else {
      count++;
    }
    std::string part = str.substr(begin, count);
    // Insert the row
    size_t packed_int = unpack(part);
    Permutation perm = permFromPosition(packed_int+minPos);
    for (size_t j = 0; j < N; j++) {
      square[i*N + j] = perm[j];
    }

    begin += count;
  }

  for (size_t col = 0; col < N; col++) {
    size_t remaining = factorial[N];
    for (size_t row = 0; row < N-1; row++) {
      remaining /= square[row*N + col];
    }
    square[(N-1)*N + col] = remaining;
  }
}
