// AUTHOR: Michael Gill
// School or Mathematics
// Monash University
// Clayton, Victoria, AU 3800
// email:  Michael.Gill@Monash.edu
//
// Implementation of the encoders. For encoder specifications see the manual io
// section
// Copyright (C) 2019  Michael Gill
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "io/encoder.hpp"
#include "util/list.hpp"

Encoder::~Encoder() {}

void NumericEncoder::encode(std::ostream &stream, const Square &square) {
  std::vector<unsigned short> numeric;
  std::transform(square.begin(), square.end(), std::back_inserter(numeric),
    [](unsigned char c) -> unsigned short { return (unsigned short) c; });

  for (size_t row = 0; row < N-1; row++) {
    stream << join(numeric.begin()+N*row, numeric.begin() + N*(row+1)) << "\n";
  }
  stream << join(numeric.begin()+N*(N-1), numeric.end());
}

void AlphaEncoder::encode(std::ostream &stream, const Square &square) {
  std::vector<unsigned char> alpha;
  std::transform(square.begin(), square.end(), std::back_inserter(alpha),
    [](unsigned char c) -> unsigned char { return c+'a'-1; });

  for (size_t row = 0; row < N-1; row++) {
    stream << join(alpha.begin()+N*row, alpha.begin() + N*(row+1), "") << "\n";
  }
  stream << join(alpha.begin()+N*(N-1), alpha.end(), "");
}

// Pack the LS contained in square into a reduced form bitpacked representation.
void BitEncoder::encode(std::ostream &stream, const Square &square) {

  // Build a map from the premutation denoted by the first row of the latin
  // square to the identity permutation.
  size_t map[N];
  for (size_t i = 0; i < N; i++) {
    map[square[i]] = i+1;
  }

  // Pack the first N-1 values of each row into the first 36-bits of an integer
  // and pack using the integer encoding.
  for (size_t i = 1; i < N-1; i++) {
    size_t value = 0;
    for (size_t j = 0; j < N-1; j++) {
      value |= (map[square[i*N+j]] << HALF_BYTE_S*(N-1-(j+1)));
    }
    stream << pack(value);
  }
}

// TODO(Micheal) before 18 May: Details as this one is complicated.
void FactorialEncoder::encode(std::ostream &stream, const Square &square) {

  Permutation derangement = getMinimalDerangement();

  Permutation map;
  for (size_t row = 0; row < N; row++) {
    map[square[row]-1] = row+1;
  }

  std::stringstream ss;
  for (size_t row = 1; row < N-1; row++) {
    stream << pack(lexPermutationPosition(square.begin() + row*N, map)
      - lexPermutationPosition(derangement.begin()));
  }
}
